/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.mindswap.pellet.owlapi.Reasoner;
import org.semanticweb.owl.inference.OWLReasoner;
import org.semanticweb.owl.inference.OWLReasonerException;
import org.semanticweb.owl.inference.OWLReasonerFactory;
import org.semanticweb.owl.model.OWLAnnotationAxiom;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLClassAxiom;
import org.semanticweb.owl.model.OWLDataProperty;
import org.semanticweb.owl.model.OWLDataPropertyAxiom;
import org.semanticweb.owl.model.OWLIndividual;
import org.semanticweb.owl.model.OWLIndividualAxiom;
import org.semanticweb.owl.model.OWLNamedObject;
import org.semanticweb.owl.model.OWLObjectProperty;
import org.semanticweb.owl.model.OWLObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyManager;

import cz.cvut.kbss.owldiff.pellet.Diff;

public class OWLAxiomTreeModel implements TreeModel {

	private static final Log LOG = LogFactory.getLog(Diff.class);

	private final List<TreeModelListener> listenerList = new ArrayList<TreeModelListener>();

	private final Collection<OWLAxiom> inferred;
	private final Collection<OWLAxiom> rest;

	private final OWLOntology ontology;
	private final OWLOntologyManager ontologyM;
	private final OWLReasonerFactory factory;

	private NodeModel<OWLOntology> rootNode;

	private OWLReasoner reasoner;

	private boolean showCommon;


	public OWLAxiomTreeModel(final DiffVisualization view,
			final Collection<OWLAxiom> uncomparable, OWLOntologyManager m,
			OWLOntology ontology, final OWLReasonerFactory f) {
		this.ontology = ontology;
		this.ontologyM = m;
		this.factory = f;

		this.inferred = new HashSet<OWLAxiom>();
		this.rest = new HashSet<OWLAxiom>(uncomparable);

		this.setView(view);
	}

	public void setInferred(final Collection<OWLAxiom> inferred) {
		this.inferred.clear();
		this.inferred.addAll(inferred);
		fireTreeNodesChanged();
	}

	private void rebuildTree(final DiffVisualization view) {
		this.rootNode = new NodeModel<OWLOntology>(ontology, null);

		final NodeModel<String> nodeIndividual;
		final NodeModel<String> nodeDataProperty;
		final NodeModel<String> nodeObjectProperty;
		final NodeModel<String> nodeClass;

		if (DiffVisualization.LIST_VIEW.equals(view)) {
			nodeClass = new NodeModel<String>("TBox Axioms", rootNode);
			nodeIndividual = new NodeModel<String>("ABox Axioms", rootNode);
			nodeDataProperty = new NodeModel<String>("RBox Axioms", rootNode);
			final NodeModel<String> other = new NodeModel<String>(
					"Other Axioms", rootNode);

			for (OWLAxiom o : ontology.getAxioms()) {
				if ((o instanceof OWLClassAxiom)) {
					nodeClass.addAxiom(o, this.rest);
				} else if (o instanceof OWLDataPropertyAxiom) {
					nodeDataProperty.addAxiom(o, this.rest);
				} else if (o instanceof OWLObjectPropertyAxiom) {
					nodeDataProperty.addAxiom(o, this.rest);
				} else if (o instanceof OWLIndividualAxiom) {
					nodeIndividual.addAxiom(o, this.rest);
				} else
					other.addAxiom(o, this.rest);
			}

			rootNode.addNodeModel(nodeClass);
			rootNode.addNodeModel(nodeIndividual);
			rootNode.addNodeModel(nodeDataProperty);
			rootNode.addNodeModel(other);
		} else {
			nodeClass = new NodeModel<String>("Classes", rootNode);
			nodeIndividual = new NodeModel<String>("Individuals", rootNode);
			nodeDataProperty = new NodeModel<String>("Data Properties",
					rootNode);
			nodeObjectProperty = new NodeModel<String>("Object Properties",
					rootNode);

			final Set<OWLAnnotationAxiom> annotations = ontology
					.getAnnotationAxioms();

			switch (view) {
			case SIMPLE_FRAME_VIEW:
				for (OWLClass ax : ontology.getReferencedClasses()) {
					final NodeModel<OWLClass> m = new NodeModel<OWLClass>(ax,
							nodeClass);
					m.addAxiom(ontology.getAxioms(ax), this.rest);
					addAnnotations(m, ax, annotations);
					nodeClass.addNodeModel(m);
				}

				for (OWLDataProperty ax : ontology
						.getReferencedDataProperties()) {
					final NodeModel<OWLDataProperty> m = new NodeModel<OWLDataProperty>(
							ax, nodeDataProperty);
					m.addAxiom(ontology.getAxioms(ax), this.rest);
					addAnnotations(m, ax, annotations);
					nodeDataProperty.addNodeModel(m);
				}

				for (OWLObjectProperty ax : ontology
						.getReferencedObjectProperties()) {
					final NodeModel<OWLObjectProperty> m = new NodeModel<OWLObjectProperty>(
							ax, nodeObjectProperty);
					m.addAxiom(ontology.getAxioms(ax), this.rest);
					addAnnotations(m, ax, annotations);
					nodeObjectProperty.addNodeModel(m);
				}

				break;
			case CLASSIFIED_FRAME_VIEW:
				try {
					reasoner = factory.createReasoner(ontologyM);
					reasoner.loadOntologies(Collections.singleton(ontology));
					if (!reasoner.isClassified()) {
						long j = System.currentTimeMillis();
						reasoner.classify();
						if (LOG.isInfoEnabled()) {
							LOG.info("Building hierarchy took "
									+ (System.currentTimeMillis() - j));
						}
					}

					final Collection<OWLClass> unsats = reasoner
							.getEquivalentClasses(ontologyM.getOWLDataFactory()
									.getOWLNothing());

					if (!unsats.isEmpty()) {
						final NodeModel<OWLClass> nm = new NodeModel<OWLClass>(
								ontologyM.getOWLDataFactory().getOWLNothing(),
								nodeClass);
						nodeClass.addNodeModel(nm);

						for (final OWLClass c : unsats) {
							final NodeModel<OWLClass> nm2 = new NodeModel<OWLClass>(
									c, nodeClass);
							nm2.addAxiom(ontology.getAxioms(c), this.rest);
							addAnnotations(nm2, c, annotations);
							nm.addNodeModel(nm2);
						}
					}

					insertAllSubNodes(nodeClass, ontologyM.getOWLDataFactory()
							.getOWLThing(), reasoner, ontology, ontologyM,
							annotations);

					for (OWLDataProperty ax : getDataTops((Reasoner) reasoner)) {
						final NodeModel<OWLDataProperty> m = new NodeModel<OWLDataProperty>(
								ax, nodeDataProperty);
						m.addAxiom(ontology.getAxioms(ax), this.rest);
						addAnnotations(m, ax, annotations);
						insertAllSubNodes(m, ax, reasoner, ontology,
								annotations);
						nodeDataProperty.addNodeModel(m);
					}

					for (OWLObjectProperty ax : getObjectTops((Reasoner) reasoner)) {
						final NodeModel<OWLObjectProperty> m = new NodeModel<OWLObjectProperty>(
								ax, nodeObjectProperty);
						m.addAxiom(ontology.getAxioms(ax), this.rest);
						addAnnotations(m, ax, annotations);
						insertAllSubNodes(m, ax, reasoner, ontology,
								annotations);
						nodeObjectProperty.addNodeModel(m);
					}
				} catch (OWLReasonerException e) {
					LOG.error(e, e);
				}
			}

			for (OWLIndividual ax : ontology.getReferencedIndividuals()) {
				NodeModel<OWLIndividual> m = new NodeModel<OWLIndividual>(ax,
						nodeIndividual);
				m.addAxiom(ontology.getAxioms(ax), this.rest);
				addAnnotations(m, ax, annotations);
				nodeIndividual.addNodeModel(m);
			}

			for (OWLAxiom ax : ontology.getAxioms()) {
				if (!nodeClass.containsAxiom(ax)
						&& !nodeDataProperty.containsAxiom(ax)
						&& !nodeObjectProperty.containsAxiom(ax)
						&& !nodeIndividual.containsAxiom(ax)) {
					rootNode.addAxiom(ax, this.rest);
				}
			}

			rootNode.addNodeModel(nodeClass);
			rootNode.addNodeModel(nodeIndividual);
			rootNode.addNodeModel(nodeDataProperty);
			rootNode.addNodeModel(nodeObjectProperty);
		}

		fireTreeStructureChanged();
	}

	private void addAnnotations(final NodeModel m, final OWLNamedObject e,
			final Collection<? extends OWLAnnotationAxiom> annotations) {
		for (OWLAnnotationAxiom ann : annotations) {
			if (ann.getSubject().equals(e)) {
				m.addAxiom(ann, rest);
			}
		}
	}

	protected void fireTreeStructureChanged() {
		final TreeModelEvent e = new TreeModelEvent(this,
				new Object[] { rootNode });
		for (final TreeModelListener l : listenerList) {
			l.treeStructureChanged(e);
		}
	}

	void fireTreeNodesChanged() {
		final TreeModelEvent e = new TreeModelEvent(this,
				new Object[] { rootNode });
		for (final TreeModelListener l : listenerList) {
			l.treeNodesChanged(e);
		}
	}

	private void insertAllSubNodes(NodeModel parent, OWLClass ax,
			OWLReasoner h, OWLOntology ontology, OWLOntologyManager m,
			Set<OWLAnnotationAxiom> annotations) throws OWLReasonerException {
		for (Set<OWLClass> xx : h.getSubClasses(ax)) {
			if (!xx.contains(m.getOWLDataFactory().getOWLNothing())) {
				for (OWLClass oc : xx) {
					final NodeModel<OWLClass> m2 = new NodeModel<OWLClass>(oc,
							parent);
					m2.addAxiom(ontology.getAxioms(oc), this.rest);
					addAnnotations(m2, oc, annotations);
					insertAllSubNodes(m2, oc, h, ontology, m, annotations);
					parent.addNodeModel(m2);
				}
			}
		}
	}

	private void insertAllSubNodes(NodeModel parent, OWLObjectProperty ax,
			OWLReasoner h, OWLOntology ontology,
			Set<OWLAnnotationAxiom> annotations) throws OWLReasonerException {
		for (Set<OWLObjectProperty> xx : h.getSubProperties(ax)) {
			for (OWLObjectProperty oc : xx) {
				NodeModel<OWLObjectProperty> m2 = new NodeModel<OWLObjectProperty>(
						oc, parent);
				m2.addAxiom(ontology.getAxioms(oc), this.rest);
				addAnnotations(m2, oc, annotations);
				insertAllSubNodes(m2, oc, h, ontology, annotations);
				parent.addNodeModel(m2);
			}
		}
	}

	private void insertAllSubNodes(NodeModel parent, OWLDataProperty ax,
			OWLReasoner h, OWLOntology ontology,
			Set<OWLAnnotationAxiom> annotations) throws OWLReasonerException {
		for (Set<OWLDataProperty> xx : h.getSubProperties(ax)) {
			for (OWLDataProperty oc : xx) {
				final NodeModel<OWLDataProperty> m2 = new NodeModel<OWLDataProperty>(
						oc, parent);
				m2.addAxiom(ontology.getAxioms(oc), this.rest);
				addAnnotations(m2, oc, annotations);
				insertAllSubNodes(m2, oc, h, ontology, annotations);
				parent.addNodeModel(m2);
			}
		}
	}

	private Collection<OWLDataProperty> getDataTops(final Reasoner r)
			throws OWLReasonerException {
		final Set<OWLDataProperty> dps = new HashSet<OWLDataProperty>();
		for (final OWLDataProperty dp : r.getDataProperties()) {
			if (r.getSuperProperties(dp).isEmpty()) {
				dps.add(dp);
			}
		}

		return dps;
	}

	private Collection<OWLObjectProperty> getObjectTops(final Reasoner r)
			throws OWLReasonerException {
		final Set<OWLObjectProperty> dps = new HashSet<OWLObjectProperty>();
		for (final OWLObjectProperty dp : r.getObjectProperties()) {
			if (r.getSuperProperties(dp).isEmpty()) {
				dps.add(dp);
			}
		}

		return dps;
	}

	public void setView(final DiffVisualization view) {
		rebuildTree(view);
	}

	public void setShowCommon(boolean showCommon) {
		this.showCommon = showCommon;
		fireTreeStructureChanged();
	}

	public boolean isCommon(final Object a) {
		return !rest.contains(a);
	}

	public boolean isInferred(final Object a) {
		return inferred.contains(a);
	}

	/**
	 * {@inheritDoc}
	 */
	public void addTreeModelListener(TreeModelListener l) {
		if (l != null && !listenerList.contains(l)) {
			listenerList.add(l);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public Object getChild(Object parent, int index) {
		return ((NodeModel) parent).getChild(index, showCommon);
	}

	/**
	 * {@inheritDoc}
	 */
	public int getChildCount(Object parent) {
		if (parent instanceof NodeModel) {
			return ((NodeModel) parent).getCount(showCommon);
		} else {
			return 0;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public int getIndexOfChild(Object parent, Object child) {
		return ((NodeModel) parent).getIndexOf(child, showCommon);
	}

	/**
	 * {@inheritDoc}
	 */
	public Object getRoot() {
		return rootNode;
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean isLeaf(Object o) {
		return (o instanceof OWLAxiom);
	}

	/**
	 * {@inheritDoc}
	 */
	public void removeTreeModelListener(TreeModelListener l) {
		if (l != null && listenerList.contains(l)) {
			listenerList.remove(l);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void valueForPathChanged(TreePath arg0, Object arg1) {
		if (LOG.isInfoEnabled()) {
			// TODO
			LOG
					.info("TreeModel is not editable, method valueForPathChanged has no effect");
		}
	}

	class NodeModel<T extends Comparable> implements Comparable<NodeModel> {

		private final T name;
		private final NodeModel<?> parent;

		// Nested Frames
		private final List<NodeModel<?>> differentNodeModels;
		private final List<NodeModel<?>> commonNodeModels;

		private final List<OWLAxiom> differentAxioms;
		private final List<OWLAxiom> commonAxioms;

		NodeModel(T name, NodeModel<?> parent) {
			this.name = name;

			this.parent = parent;

			this.commonNodeModels = new ArrayList<NodeModel<?>>();
			this.differentNodeModels = new ArrayList<NodeModel<?>>();

			this.commonAxioms = new ArrayList<OWLAxiom>();
			this.differentAxioms = new ArrayList<OWLAxiom>();
		}

		public void addNodeModel(NodeModel<?> n) {
			if (n.containsDifferent() || n.containsInferred()) {
				this.differentNodeModels.add(n);
				if (parent != null) {
					parent.putToDifferentInParent(this);
				}
			} else {
				this.commonNodeModels.add(-Collections.binarySearch(
						this.commonNodeModels, n) - 1, n);
			}
		}

		// Due to hierarchy, it is necessary to check all parents to see if they
		// are
		// correctly marked as containing different axioms
		private void putToDifferentInParent(NodeModel<?> m) {
			if (commonNodeModels.contains(m)) {
				commonNodeModels.remove(m);
				this.differentNodeModels.add(-Collections.binarySearch(
						this.differentNodeModels, m) - 1, m);
			}
			if (parent != null) {
				parent.putToDifferentInParent(this);
			}
		}

		public void addAxiom(Collection<? extends OWLAxiom> c,
				Collection<? extends OWLAxiom> common) {
			for (OWLAxiom a : c) {
				addAxiom(a, common);
			}
		}

		public void addAxiom(OWLAxiom c, Collection<? extends OWLAxiom> diff) {
			if (diff.contains(c)) {
				this.differentAxioms.add(-Collections.binarySearch(
						this.differentAxioms, c) - 1, c);
			} else {
				this.commonAxioms.add(-Collections.binarySearch(
						this.commonAxioms, c) - 1, c);
			}
		}

		public Object getObject() {
			return this.name;
		}

		public int getCount(boolean showCommon) {
			if (showCommon) {
				return this.commonAxioms.size() + this.differentAxioms.size()
						+ this.commonNodeModels.size()
						+ this.differentNodeModels.size();
			} else {
				return this.differentAxioms.size()
						+ this.differentNodeModels.size();
			}
		}

		public int getIndexOf(Object o, boolean showCommon) {
			int index = -1;

			if (differentAxioms.contains(o)) {
				return differentAxioms.indexOf(o);
			}

			index += differentAxioms.size();
			if (showCommon) {
				if (commonAxioms.contains(o)) {
					return index + commonAxioms.indexOf(o);
				}
				index += commonAxioms.size();
			}

			if (differentNodeModels.contains(o)) {
				return index + differentNodeModels.indexOf(o);
			}
			index += differentNodeModels.size();

			return index + commonNodeModels.indexOf(o);
		}

		boolean containsInferred() {
			if (!Collections.disjoint(inferred, differentAxioms)) {
				return true;
			}

			if (!differentNodeModels.isEmpty()) {
				for (NodeModel<?> node : differentNodeModels) {
					if (node.containsInferred()) {
						return true;
					}
				}
			}
			if (!commonNodeModels.isEmpty()) {

				for (NodeModel<?> node : commonNodeModels) {
					if (node.containsInferred()) {
						return true;
					}
				}

			}
			return false;
		}

		boolean containsDifferent() {
			if (!Collections.disjoint(rest, differentAxioms)) {
				return true;
			}
			if (!differentNodeModels.isEmpty()) {
				for (NodeModel<?> node : differentNodeModels) {
					if (node.containsDifferent()) {
						return true;
					}
				}
			}
			if (!commonNodeModels.isEmpty()) {
				for (NodeModel<?> node : commonNodeModels) {
					if (node.containsDifferent()) {
						return true;
					}
				}
			}
			return false;
		}

		boolean containsAxiom(OWLAxiom ax) {
			if (this.commonAxioms.contains(ax)
					|| this.differentAxioms.contains(ax)) {
				return true;
			}
			if (!commonNodeModels.isEmpty()) {
				for (NodeModel<?> node : commonNodeModels) {
					if (node.containsAxiom(ax)) {
						return true;
					}
				}
			}
			if (!differentNodeModels.isEmpty()) {
				for (NodeModel<?> node : differentNodeModels) {
					if (node.containsAxiom(ax)) {
						return true;
					}
				}
			}
			return false;
		}

		Object getChild(int index, boolean showCommon) {
			if (index < differentAxioms.size()) {
				return differentAxioms.get(index);
			}
			index = index - differentAxioms.size();

			if (showCommon) {
				if (index < commonAxioms.size()) {
					return commonAxioms.get(index);
				}
				index = index - commonAxioms.size();
			}

			if (index < this.differentNodeModels.size()) {
				return differentNodeModels.get(index);
			}
			index = index - differentNodeModels.size();

			// only if showCommon is set
			return commonNodeModels.get(index);
		}

		public int compareTo(NodeModel o) {
			return name.compareTo(o.name);
		}

		@Override
		public int hashCode() {
			return name.hashCode();
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			NodeModel other = (NodeModel) obj;
			if (name == null) {
				if (other.name != null)
					return false;
			} else if (!name.equals(other.name))
				return false;
			return true;
		}

		public String toString() {
			return name.toString();
		}
	}
}
