/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff.cex;

import java.io.File;
import java.net.URI;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLDataFactory;
import org.semanticweb.owl.model.OWLObjectProperty;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyCreationException;
import org.semanticweb.owl.model.OWLOntologyManager;

import cz.cvut.kbss.owldiff.DLSyntax;
import cz.cvut.kbss.owldiff.OWLDiffException;
import cz.cvut.kbss.owldiff.ProgressListener;

public class Diff  {

	private static final Log LOG = LogFactory.getLog(Diff.class);

	private static final int PROGRESS_MAX = 5;

	private OWLOntologyManager manager;

	private OWLDataFactory factory;

	private Set<OWLClass> diffL, diffR;

	private ProgressListener progressListener;

	public Diff(ProgressListener progressListener) {
		this.progressListener = progressListener;
	}

	/**
	 * {@inheritDoc}
	 */
	public void diff(final URI o, final URI u) throws OWLDiffException {
		diff(o, u, false);
	}

	/**
	 * Convenience method when parsing from scratch.
	 */
	public void diff(final URI o, final URI u, boolean askForSigma)
			throws OWLDiffException {

		try {
			OWLOntologyManager originalM = org.semanticweb.owl.apibinding.OWLManager
					.createOWLOntologyManager();
			OWLOntology original = originalM.loadOntology(new File(o).toURI());

			OWLOntologyManager updateM = org.semanticweb.owl.apibinding.OWLManager
					.createOWLOntologyManager();
			OWLOntology update = updateM.loadOntology(new File(u).toURI());

			if (LOG.isDebugEnabled()) {
				LOG.debug("Original: " + axiomsToString(original.getAxioms()));
			}

			if (LOG.isDebugEnabled()) {
				LOG.debug("Update: " + axiomsToString(update.getAxioms()));
			}

			diff(originalM, original, updateM, update, askForSigma);
		} catch (OWLOntologyCreationException e1) {
			LOG.error(e1, e1);
		}

	}

	public void diff(final OWLOntologyManager om, OWLOntology original,
			final OWLOntologyManager um, OWLOntology update)
			throws OWLDiffException {
		diff(om, original, um, update, false);
	}

	public void diff(final OWLOntologyManager om, OWLOntology original,
			final OWLOntologyManager um, OWLOntology update, boolean askForSigma)
			throws OWLDiffException {

		if (progressListener != null) {
			progressListener.setProgress(0);
			progressListener.setProgressMax(PROGRESS_MAX);
		}

		// NORMALIZATION
		Normalizator origNorm = new Normalizator(original, progressListener);
		original = origNorm.normalizeTerminology();
		if (LOG.isDebugEnabled()) {
			LOG.debug("Original normalized: "
					+ axiomsToString(original.getAxioms()));
		}

		Normalizator updateNorm = new Normalizator(update, progressListener);
		update = updateNorm.normalizeTerminology();
		if (LOG.isDebugEnabled()) {
			LOG.debug("Update normalized: "
					+ axiomsToString(update.getAxioms()));
		}

		if ((original == null) || (update == null)) {
			LOG.error("Failed to normalize an input ontology");
			return;
		}

		// CEX DIFF
		Sig origSig = new Sig(original);
		Sig updateSig = new Sig(update);
		Sig sigma = createFullSigma(origSig, updateSig);

		if (LOG.isDebugEnabled()) {
			LOG.debug("Full sigma: " + sigma.toString());
		}
		if (askForSigma) {
			SigmaFrame sf = new SigmaFrame(sigma);
			sf.setVisible(true);
		}
		if (LOG.isInfoEnabled()) {
			LOG.info("Selected sigma: " + sigma.toString());
		}

		manager = OWLManager.createOWLOntologyManager();
		factory = manager.getOWLDataFactory();
		Noimply nio = new Noimply(manager, original, origSig, factory, sigma);
		nio.nnoImply();
		manager = OWLManager.createOWLOntologyManager();
		factory = manager.getOWLDataFactory();
		Noimply niu = new Noimply(manager, update, updateSig, factory, sigma);
		niu.nnoImply();

		if (progressListener != null) {
			progressListener.setProgress(1);
		}

		if (LOG.isDebugEnabled()) {
			for (OWLClass a : origSig.getSig()) {
				String name = a.getURI().getFragment();
				Set<OWLClass> n = nio.noImply(a);
				Set<OWLAxiom> nn = nio.nnoImply(a);
				Set<OWLClass> pre = nio.pre(a);
				LOG.debug("orig: " + name + ": noImply: " + classesToString(n));
				LOG.debug("orig: " + name + ": NoImply: " + axiomsToString(nn));
				LOG.debug("orig: " + name + ": pre: " + classesToString(pre));
			}
			LOG.debug("Xi orig: " + classesToString(nio.getXi()));
			LOG.info("NoImply original: " + axiomsToString(nio.nnoImply()));

			for (OWLClass a : updateSig.getSig()) {
				String name = a.getURI().getFragment();
				Set<OWLClass> n = niu.noImply(a);
				Set<OWLAxiom> nn = niu.nnoImply(a);
				Set<OWLClass> pre = niu.pre(a);
				LOG.debug("update: " + name + ": noImply: "
						+ classesToString(n));
				LOG.debug("update: " + name + ": NoImply: "
						+ axiomsToString(nn));
				LOG.debug("update: " + name + ": pre: " + classesToString(pre));
			}
			LOG.debug("Xi update: " + classesToString(niu.getXi()));
			LOG.info("NoImply update: " + axiomsToString(niu.nnoImply()));
		}

		// Result result = new Result();

		DiffR dR = new DiffR(original, update, origSig, updateSig, sigma, nio,
				niu, manager);
		dR.mark();
		if (progressListener != null) {
			progressListener.setProgress(2);
		}
		// result.diffR = dR.diffR();
		diffR = dR.diffR();
		if (progressListener != null) {
			progressListener.setProgress(3);
		}
		if (LOG.isInfoEnabled()) {
			LOG.info("DiffR: " + classesToString(diffR));
		}

		DiffL dL = new DiffL(original, update, origSig, updateSig, sigma, nio,
				niu);
		dL.generateOs();
		if (progressListener != null) {
			progressListener.setProgress(4);
		}
		if (LOG.isDebugEnabled()) {
			StringBuffer sb = new StringBuffer();
			sb.append("O: ");
			for (DiffL.OWLClassPair p : dL.getO()) {
				sb.append(p + ",");
			}
			LOG.debug(sb.toString());
		}
		// result.diffL = dL.diffL();
		diffL = dL.diffL();
		if (progressListener != null) {
			progressListener.setProgress(5);
		}
		if (LOG.isInfoEnabled()) {
			LOG.info("DiffL: " + classesToString(diffL));
		}

		// final ExplanationManager<OWLAxiom> origEM =
		// getExplanationManager(original);
		// Reasoner origR = new Reasoner(originalM);
		// origR.loadOntology(original);

		// return result;
	}

	/*
	 * public class Result { public Set<OWLClass> diffL; public Set<OWLClass>
	 * diffR; }
	 */

	private Sig createFullSigma(Sig origSig, Sig updateSig) {
		Set<OWLClass> sigo = origSig.getSig();
		Set<OWLClass> sigu = updateSig.getSig();
		Set<OWLClass> sigmaSet = new HashSet<OWLClass>();
		for (OWLClass c : sigo) {
			if (sigu.remove(c)) {
				sigmaSet.add(c);
			}
		}

		Set<OWLObjectProperty> roleso = origSig.getRoles();
		Set<OWLObjectProperty> rolesu = updateSig.getRoles();
		Set<OWLObjectProperty> roleSet = new HashSet<OWLObjectProperty>();
		for (OWLObjectProperty p : roleso) {
			if (rolesu.remove(p)) {
				roleSet.add(p);
			}
		}

		return new Sig(sigmaSet, roleSet);
	}

	public Set<OWLClass> getDiffL() {
		return diffL;
	}

	public Set<OWLClass> getDiffR() {
		return diffR;
	}

	static String classesToString(Set<OWLClass> set) {
		StringBuffer sb = new StringBuffer();
		if (set == null) {
			return sb.toString();
		}
		for (OWLClass c : set) {
			sb.append(c.getURI().getFragment() + ",");
		}
		return sb.toString();
	}

	private static String axiomsToString(Set<OWLAxiom> set) {
		StringBuffer sb = new StringBuffer();
		if (set == null) {
			return sb.toString();
		}

		for (final OWLAxiom a : set) {
			sb.append(new DLSyntax().writeAxiom(a, false, null, false));
			sb.append(", ");
		}

		return sb.toString();
	}
}
