/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff.pellet;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLImportsDeclaration;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyAnnotationAxiom;
import org.semanticweb.owl.model.OWLOntologyManager;

import cz.cvut.kbss.owldiff.OWLDiffException;
import cz.cvut.kbss.owldiff.OWLEntailmentChecker;
import cz.cvut.kbss.owldiff.ProgressListener;
import cz.cvut.kbss.owldiff.OWLDiffException.Reason;

public class Diff {

	private static final Log LOG = LogFactory.getLog(Diff.class);

	private ProgressListener progressListener;

	private OWLOntologyManager originalM, updateM;

	private OWLOntology original, update;

	private OWLEntailmentChecker originalR, updateR;

	// syntactic diff
	private final Set<OWLAxiom> origRest = new HashSet<OWLAxiom>();

	private final Set<OWLAxiom> updateRest = new HashSet<OWLAxiom>();

	// entailment
	private final Set<OWLAxiom> inferred = new HashSet<OWLAxiom>();

	private final Set<OWLAxiom> possiblyRemove = new HashSet<OWLAxiom>();

	private boolean isEntailmentComputed;

	public Diff(ProgressListener progressListener) {
		this.progressListener = progressListener;
	}

	/**
	 * {inheritDoc}
	 */
	public void diff(final OWLOntologyManager om, final OWLOntology o,
			final OWLOntologyManager um, final OWLOntology u)
			throws OWLDiffException {
		diff(om, o, um, u, true);
	}

	public void diff(final OWLOntologyManager originalM,
			final OWLOntology original, final OWLOntologyManager updateM,
			final OWLOntology update, boolean entailments)
			throws OWLDiffException {

		this.originalM = originalM;
		this.updateM = updateM;
		this.update = update;
		this.original = original;

		this.originalR = null;
		this.updateR = null;

		if (entailments) {
			computeEntailments();
		} else {
			syntacticDiff();
		}
	}

	private OWLEntailmentChecker create(final OWLOntologyManager m,
			final OWLOntology o) {
		return new PelletEntailmentChecker(m, o);
	}

	private void computeEntailments() throws OWLDiffException {
		if (originalR == null) {
			originalR = create(originalM, original);
		}

		if (!originalR.isConsistent()) {
			throw new OWLDiffException(Reason.INCONSISTENT_ONTOLOGY,
					"Original Ontology is inconsistent.");
		}

		int i = 0;
		if (progressListener != null) {
			progressListener
					.setProgressMax(origRest.size() + updateRest.size());
		}

		for (final OWLAxiom a : updateRest) {
			updateProgress(++i);

			if (a.isLogicalAxiom()) {
				if (originalR.isEntailed(a)) {
					possiblyRemove.add(a);
				}
			}
		}

		if (LOG.isInfoEnabled()) {
			LOG.info("Update Inferred: " + possiblyRemove);
		}

		if (updateR == null) {
			updateR = create(updateM, update);
		}

		if (!updateR.isConsistent()) {
			throw new OWLDiffException(Reason.INCONSISTENT_ONTOLOGY,
					"Update Ontology is inconsistent.");
		}

		for (final OWLAxiom a : origRest) {
			updateProgress(++i);

			if (a.isLogicalAxiom()) {
				if (updateR.isEntailed(a)) {
					inferred.add(a);
				}
			}
		}

		if (LOG.isInfoEnabled()) {
			LOG.info("Original Inferred: " + inferred);
		}

		isEntailmentComputed = true;
	}

	private void syntacticDiff() {
		updateRest.clear();
		origRest.clear();

		isEntailmentComputed = false;
		inferred.clear();
		possiblyRemove.clear();

		if (LOG.isInfoEnabled()) {
			LOG.info("Original # of all axioms=" + original.getAxioms().size()
					+ ", # of logical axioms="
					+ original.getLogicalAxioms().size());
			LOG.info("Update # of all axioms=" + update.getAxioms().size()
					+ ", # of logical axioms="
					+ update.getLogicalAxioms().size());
		}

		int i = 0;
		if (progressListener != null) {
			progressListener.setProgressMax(update.getAxioms().size()
					+ original.getAxioms().size());
		}

		for (final OWLAxiom a : update.getAxioms()) {
			updateProgress(++i);

			if (original.containsAxiom(a)) {
				continue;
			}

			if (a instanceof OWLOntologyAnnotationAxiom) {
				if (!containsWithSameOntologyURI(
						(OWLOntologyAnnotationAxiom) a, original
								.getOntologyAnnotationAxioms())) {
					updateRest.add(a);
				}
			} else if (a instanceof OWLImportsDeclaration) {
				if (!containsWithSameOntologyURI((OWLImportsDeclaration) a,
						original.getImportsDeclarations())) {
					updateRest.add(a);
				}
			} else {
				updateRest.add(a);
			}
		}

		for (final OWLAxiom a : original.getAxioms()) {
			updateProgress(++i);

			if (update.containsAxiom(a)) {
				continue;
			}

			if (a instanceof OWLOntologyAnnotationAxiom) {
				if (!containsWithSameOntologyURI(
						(OWLOntologyAnnotationAxiom) a, update
								.getOntologyAnnotationAxioms())) {
					origRest.add(a);
				}
			} else if (a instanceof OWLImportsDeclaration) {
				if (!containsWithSameOntologyURI((OWLImportsDeclaration) a,
						update.getImportsDeclarations())) {
					origRest.add(a);
				}
			} else {
				origRest.add(a);
			}
		}

		if (LOG.isInfoEnabled()) {
			LOG
					.info("========================================================================");
			LOG.info("Original Rest: " + origRest);
			LOG.info("Update Rest: " + updateRest);
		}
	}

	private boolean containsWithSameOntologyURI(
			final OWLOntologyAnnotationAxiom o,
			final Set<OWLOntologyAnnotationAxiom> a) {
		for (final OWLOntologyAnnotationAxiom ax : a) {
			if (equalsWithRespectToOntologyURI(o, ax)) {
				return true;
			}
		}
		return false;
	}

	private boolean containsWithSameOntologyURI(final OWLImportsDeclaration o,
			final Set<OWLImportsDeclaration> a) {
		for (final OWLImportsDeclaration ax : a) {
			if (equalsWithRespectToOntologyURI(o, ax)) {
				return true;
			}
		}
		return false;
	}

	private boolean equalsWithRespectToOntologyURI(
			final OWLOntologyAnnotationAxiom o,
			final OWLOntologyAnnotationAxiom o2) {
		if (o.equals(o2)) {
			return true;
		}

		return o.getAnnotation().equals(o2.getAnnotation())
				&& o.getSubject().getURI().equals(o2.getSubject().getURI());
	}

	private boolean equalsWithRespectToOntologyURI(
			final OWLImportsDeclaration o, final OWLImportsDeclaration o2) {
		if (o.equals(o2)) {
			return true;
		}

		return o.getImportedOntologyURI().equals(o2.getImportedOntologyURI())
				&& o.getSubject().getURI().equals(o2.getSubject().getURI());
	}

	private void updateProgress(int i) {
		if (progressListener != null) {
			progressListener.setProgress(i);
		}
	}

	public Collection<OWLAxiom> getUpdateDiff() {
		return updateRest;
	}

	public Collection<OWLAxiom> getOriginalDiff() {
		return origRest;
	}

	public Collection<OWLAxiom> getOriginalInferred() {
		return inferred;
	}

	public Collection<OWLAxiom> getUpdateInferred() {
		return possiblyRemove;
	}

	public boolean isEntailmentComputed() {
		return isEntailmentComputed;
	}
}
